
import sys
import os
import codecs

import kindred

def saveDataFromSTFormat(data,predictedRelations,txtPath,a1Path,a2Path):
	assert isinstance(data,kindred.Document)

	with codecs.open(txtPath,'w','utf8') as txtFile, codecs.open(a1Path,'w','utf8') as a1File, codecs.open(a2Path,'w','utf8') as a2File:
		txtFile.write(data.getText())
		
		for e in data.getEntities():
			assert isinstance(e,kindred.Entity)
		
			positions = ";".join("%d %d" % (start,end) for start,end in e.position)
			line = "%s\t%s %s\t%s" % (e.sourceEntityID,e.entityType,positions,e.text)
			a1File.write(line+"\n")
			
		entityIDsToSourceEntityIDs = data.getEntityIDsToSourceEntityIDs()	
		
		for i,r in enumerate(data.getRelations()):
			assert isinstance(r,kindred.Relation)
			
			relationType = r.relationType
			relationEntityIDs = [ entityIDsToSourceEntityIDs[eID] for eID in r.entityIDs ]
			
			if r.argNames is None:
				argNames = [ ("arg%d" % (i+1)) for i in range(len(relationEntityIDs)) ]
			else:
				argNames = r.argNames

			arguments = " ".join(["%s:%s" % (a,b) for a,b in zip(argNames,relationEntityIDs) ])
			line = "R%d\t%s %s" % (i,relationType,arguments)
			a2File.write(line+"\n")
			
		for i,r in enumerate(predictedRelations):
			assert isinstance(r,kindred.Relation)
			
			relationType = r.relationType
			inThisText = [ eID in entityIDsToSourceEntityIDs for eID in r.entityIDs ]
			if not all(inThisText):
				# Not all the entities in this relation are in this portion of text. So don't include
				continue
			
			relationEntityIDs = [ entityIDsToSourceEntityIDs[eID] for eID in r.entityIDs ]
			
			if r.argNames is None:
				argNames = [ ("arg%d" % (i+1)) for i in range(len(relationEntityIDs)) ]
			else:
				argNames = r.argNames

			arguments = " ".join(["%s:%s" % (a,b) for a,b in zip(argNames,relationEntityIDs) ])
			line = "R%d\t%s %s" % (i,relationType,arguments)
			a2File.write(line+"\n")
			


def save(corpus,dataFormat,directory,predictedRelations=[]):
	"""
	Save a corpus to a directory
	"""
	
	assert dataFormat == 'standoff' or dataFormat == 'simpletag' or dataFormat == 'json'

	assert isinstance(corpus,kindred.Corpus)
	
	for i,d in enumerate(corpus.documents):
		if dataFormat == 'standoff':
			
			if d.getSourceFilename() is None:
				base = "%08d" % i
			else:
				base = d.getSourceFilename()

			txtPath = os.path.join(directory,'%s.txt' % base)
			a1Path = os.path.join(directory,'%s.a1' % base)
			a2Path = os.path.join(directory,'%s.a2' % base)

			saveDataFromSTFormat(d,predictedRelations,txtPath,a1Path,a2Path)

	



import kindred


class Relation:
	"""
	Describes relationship between entities (including relation type and argument names if applicable).
	"""
	
	def __init__(self,relationType=None,entityIDs=[],argNames=None):
		assert isinstance(entityIDs,list)

		self.relationType = relationType
		self.entityIDs = entityIDs
		if argNames == None:
			self.argNames = None
		else:
			self.argNames = [ str(a) for a in argNames ]
	
	def __eq__(self, other):
		"""Override the default Equals behavior"""
		if isinstance(other, self.__class__):
			return self.__dict__ == other.__dict__
		return False
	
	def __ne__(self, other):
		"""Define a non-equality test"""
		return not self.__eq__(other)

	def __str__(self):
		return "<Relation %s %s %s>" % (self.relationType,str(self.entityIDs),str(self.argNames))

	def __repr__(self):
		return self.__str__()

	def __hash__(self):
		if self.argNames is None:
			return hash((self.relationType,tuple(self.entityIDs)))
		else:
			return hash((self.relationType,tuple(self.entityIDs),tuple(self.argNames)))




import sys

from sklearn import svm
import numpy as np
from collections import defaultdict

from sklearn.feature_selection import SelectKBest,chi2,SelectPercentile,RFECV
from sklearn.metrics import f1_score,confusion_matrix
from sklearn.ensemble import AdaBoostClassifier,GradientBoostingClassifier,BaggingClassifier,RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier

from sklearn.model_selection import cross_val_score,cross_val_predict,StratifiedKFold
from sklearn.pipeline import Pipeline
from sklearn.svm import LinearSVC,SVC

from scipy.sparse import coo_matrix, csr_matrix, lil_matrix, hstack, vstack

import kindred
from kindred.CandidateBuilder import CandidateBuilder
from kindred.Vectorizer import Vectorizer

class RelationClassifier:
	"""
	Manages binary classifier(s) for relation classification.
	"""
	def __init__(self,useSingleClassifier=True,useBuilder=False,tfidf=True,features=None,threshold=None):
		"""
		Constructor for the RelationClassifier class
		
		:param self: Object instance
		:param useSingleClassifier: Whether to use a single classifier
		:param useBuilder: Whether to use the feature builder functionality
		:param tfidf: Whether to use tfidf for the vectorizer
		:param features: A list of specific features
		:param threshold: A specific threshold to use for classification (which will then use a logistic regression classifier)
		"""
		self.isTrained = False
		self.useSingleClassifier = useSingleClassifier
		self.useBuilder = useBuilder
		self.tfidf = tfidf

		self.defaultFeatures = ["entityTypes","unigramsBetweenEntities","bigrams","dependencyPathEdges","dependencyPathEdgesNearEntities"]
		if not features is None:
			assert isinstance(features,list)
			self.defaultFeatures = features
			
		self.threshold = threshold
		#self.defaultFeatures = ["entityTypes","dependencyPathEdges"]

	def buildFeatureSet(self,corpus,candidateRelations,classes,tfidf):
		"""
		Builds the set of features that gives best cross-validated F1-score
		
		:param self: Object instance
		:param corpus: Corpus of documents to use for training
		:param candidateRelations: List of candidate relations to use for training
		:param classes: Associated numerical classes for training
		:param tfidf: Whether to use tfidf for the vectorizer
		:returns: A list of features to use for vectorizer
		"""
		
		vectorizers = {}
		trainVectors = {}

		featureChoice = ["entityTypes","unigramsBetweenEntities","bigrams","dependencyPathEdges","dependencyPathEdgesNearEntities"]
		for feature in featureChoice:
			vectorizers[feature] = Vectorizer(featureChoice=[feature],tfidf=tfidf)
			trainVectors[feature] = vectorizers[feature].fit_transform(corpus)

		groupVector = None
		chosenFeatures = []
		prevScore,prevMatrix = -1.0, None
		while True:
			bestScore, bestFeature, bestMatrix = -1.0, None, None
			for feature in featureChoice:
				if prevMatrix is None:
					matrix = trainVectors[feature]
				else:
					matrix = hstack([prevMatrix,trainVectors[feature]])

				clf = svm.LinearSVC(class_weight='balanced')
				
				scores = cross_val_score(clf, matrix, classes, cv=5, scoring='f1_macro')
				score = scores.mean()

				if score > bestScore:
					bestScore = score
					bestFeature = feature
					bestMatrix = matrix

			if bestScore > prevScore:
				# We see improvement
				featureChoice.remove(bestFeature)
				chosenFeatures.append(bestFeature)
				prevScore = bestScore
				prevMatrix = bestMatrix
			else:
				# No improvement made
				break

		return chosenFeatures

	def train(self,corpus):
		"""
		Does stuff
		"""
		assert isinstance(corpus,kindred.Corpus)
			
		self.candidateBuilder = CandidateBuilder()
		self.candidateBuilder.fit_transform(corpus)
		
		self.relTypeToValidEntityTypes = defaultdict(set)
		
		for d in corpus.documents:
			for r in d.getRelations():
				entityIDsToEntities = d.getEntityIDsToEntities()
				relationEntities = [ entityIDsToEntities[eID] for eID in r.entityIDs ]
				validEntityTypes = tuple([ e.entityType for e in relationEntities ])
				
				relKey = tuple([r.relationType] + r.argNames)
				self.relTypeToValidEntityTypes[relKey].add(validEntityTypes)
			
				
		self.classToRelType = { (i+1):relType for i,relType in enumerate(corpus.relationTypes) }

		# Get the set of valid classes
		relationtypeCount = len(corpus.relationTypes)
		allClasses = list(range(1,relationtypeCount+1))
		self.allClasses = allClasses
	
		#options = ["ngrams","selectedngrams","bigrams","ngramsPOS","selectedngramsPOS","ngramsOfDependencyPath","bigramsOfDependencyPath","entityTypes","lemmas","selectedlemmas","dependencyPathEdges","dependencyPathEdgesNearEntities","splitAcrossSentences","skipgrams_2","skipgrams_3","skipgrams_4","skipgrams_5","skipgrams_6","skipgrams_7","skipgrams_8","skipgrams_9","skipgrams_10","unigramsBetweenEntities","bigrams_betweenEntities"]

		# We'll just get the vectors for the entityTypes

		#tmpClassData = [ (1 in candidateClassGroup) for candidateClassGroup in candidateClasses ]

		candidateRelations = corpus.getCandidateRelations()
		candidateClasses = corpus.getCandidateClasses()

		#useSingleClassifier = False
		if self.useSingleClassifier:
			#chosenFeatures = ["entityTypes","dependencyPathEdges","unigramsBetweenEntities","bigrams_betweenEntities","bigramsOfDependencyPath"]

			simplifiedClasses = []
			# TODO: Try sparse matrix rep
			for candidateRelation,candidateClassGroup in zip(candidateRelations,candidateClasses):
				#assert len(candidateClassGroup) == 1, "Cannot handle multiple relations with same set of entities " + str(candidateRelation)
				simplifiedClasses.append(candidateClassGroup[0])
			#assert False
	
			if self.useBuilder:
				chosenFeatures = self.buildFeatureSet(corpus,candidateRelations,simplifiedClasses,self.tfidf)
			else:
				chosenFeatures = self.defaultFeatures

			self.vectorizer = Vectorizer(featureChoice=chosenFeatures,tfidf=self.tfidf)
			trainVectors = self.vectorizer.fit_transform(corpus)
		
			assert trainVectors.shape[0] == len(candidateClasses)
		
			if self.threshold is None:
				self.clf = svm.LinearSVC(class_weight='balanced',random_state=1)
			else:
				self.clf = kindred.LogisticRegressionWithThreshold(self.threshold)
			self.clf.fit(trainVectors,simplifiedClasses)
		else:
			# TODO: Should we take into account the argument count when grouping classifiers?

			if not self.useBuilder:
				chosenFeatures = self.defaultFeatures

				self.vectorizer = Vectorizer(featureChoice=chosenFeatures,tfidf=self.tfidf)
				tmpMatrix = self.vectorizer.fit_transform(corpus)
			self.clfs = {}
			self.vectorizers = {}
			for c in self.allClasses:
				tmpClassData = [ (c in candidateClassGroup) for candidateClassGroup in candidateClasses ]

				if self.useBuilder:
					chosenFeatures = self.buildFeatureSet(corpus,candidateRelations,tmpClassData,self.tfidf)
					self.vectorizers[c] = Vectorizer(featureChoice=chosenFeatures,tfidf=self.tfidf)
					tmpMatrix = self.vectorizers[c].fit_transform(corpus)

				#save_sparse_csr('train.matrix',trainVectors.tocsr())
				#saveClasses('train.classes',tmpClassData)

				if self.threshold is None:
					self.clfs[c] = svm.LinearSVC(class_weight='balanced',random_state=1)
				else:
					self.clfs[c] = kindred.LogisticRegressionWithThreshold(self.threshold)
				#self.clfs[c] = AdaBoostClassifier(n_estimators=2)
				#self.clfs[c].fit(trainVectors,tmpClassData)
				self.clfs[c].fit(tmpMatrix,tmpClassData)
		
		self.isTrained = True

	def predict(self,corpus):
		assert self.isTrained, "Classifier must be trained using train() before predictions can be made"
	
		assert isinstance(corpus,kindred.Corpus)
			
		self.candidateBuilder.transform(corpus)

		#if False:
		#	testVectors = self.vectorizer.transform(candidateRelations)
		#	tmpClassData = [ (1 in candidateClassGroup) for candidateClassGroup in testClasses ]
		
		
		#save_sparse_csr('test.matrix',testVectors.tocsr())
		#saveClasses('test.classes',tmpClassData)

		candidateRelations = corpus.getCandidateRelations()
		candidateClasses = corpus.getCandidateClasses()
		
		entityIDsToType = {}
		for doc in corpus.documents:
			for e in doc.getEntities():
				entityIDsToType[e.entityID] = e.entityType
		
		predictedRelations = []
		
		if self.useSingleClassifier:
			tmpMatrix = self.vectorizer.transform(corpus)

			#predictedClasses = self.clfs[c].predict(testVectors)
			predictedClasses = self.clf.predict(tmpMatrix)
			for predictedClass,candidateRelation in zip(predictedClasses,candidateRelations):
				if predictedClass != 0:
					relKey = self.classToRelType[predictedClass]
					relType = relKey[0]
					argNames = relKey[1:]
					
					candidateRelationEntityTypes = tuple( [ entityIDsToType[eID] for eID in candidateRelation.entityIDs ] )
					if not tuple(candidateRelationEntityTypes) in self.relTypeToValidEntityTypes[relKey]:
						continue

					predictedRelation = kindred.Relation(relType,candidateRelation.entityIDs,argNames=argNames)
					predictedRelations.append(predictedRelation)
		else:
			if not self.useBuilder:
				tmpMatrix = self.vectorizer.transform(corpus)

			for c in self.allClasses:

				if self.useBuilder:
					tmpMatrix = self.vectorizers[c].transform(corpus)

				#predictedClasses = self.clfs[c].predict(testVectors)
				predicted = self.clfs[c].predict(tmpMatrix)
				for p,candidateRelation in zip(predicted,candidateRelations):
					if p != 0:
						relKey = self.classToRelType[p]
						relType = relKey[0]
						argNames = relKey[1:]
						
						candidateRelationEntityTypes = tuple( [ entityIDsToType[eID] for eID in candidateRelation.entityIDs ] )
						if not tuple(candidateRelationEntityTypes) in self.relTypeToValidEntityTypes[relKey]:
							continue
						
						predictedRelation = kindred.Relation(relType,candidateRelation.entityIDs,argNames=argNames)
						predictedRelations.append(predictedRelation)
		
		# Add the predicted relations into the corpus
		entitiesToDoc = {}
		for i,doc in enumerate(corpus.documents):
			for e in doc.getEntities():
				entitiesToDoc[e.entityID] = i

		for predictedRelation in predictedRelations:
			docIDs = [ entitiesToDoc[eID] for eID in predictedRelation.entityIDs ]
			docIDs = list(set(docIDs))
			assert len(docIDs) > 0, "Predicted relation contains entities that don't match any documents in corpus"
			assert len(docIDs) == 1, "Predicted relation contains entities that are spread across documents"

			docID = docIDs[0]
			corpus.documents[docID].addRelation(predictedRelation)




import kindred

import hashlib

# random module is not consistent across platforms
# Based on idea from:
# https://stackoverflow.com/questions/9023660/how-to-generate-a-repeatable-random-number-sequence
hashVal = b'seed'
def customSeed(seed):
	global hashVal
	assert isinstance(seed,bytes)
	hashVal = seed

def customRandom():
	global hashVal
	hashVal = hashlib.md5(hashVal).digest()
	
	if isinstance(hashVal,str):
		num = 256*ord(hashVal[0]) + ord(hashVal[1])
	else:
		num = 256*hashVal[0] + hashVal[1]
		
	return num/float(256*256)

def customChoice(seq):
	index = int(customRandom() * len(seq))
	return seq[index]

def customSample(population,k):
	listclone = list(population)
	assert k <= len(population)
	chosen = []
	for i in range(k):
		choice = int(customRandom() * len(listclone) )
		item = listclone.pop(choice)
		chosen.append(item)
	return chosen

def generateData(positiveCount=100,negativeCount=100,relTypes=1):
	customSeed(b'seed')
	positivePatterns = ['<drug id="ID1">DRUG</drug> treats <disease id="ID2">DISEASE</disease>.',
						'<drug id="ID1">DRUG</drug> is a common treatment for <disease id="ID2">DISEASE</disease>.',
						'<drug id="ID1">DRUG</drug> is often used for <disease id="ID2">DISEASE</disease>.',
						'<disease id="ID2">DISEASE</disease> can be treated with <drug id="ID1">DRUG</drug>.']
	negativePatterns = ['<drug id="ID1">DRUG</drug> and <disease2 id="ID2">DISEASE</disease2> were discovered by the same researcher.',
						'<drug id="ID1">DRUG</drug> is the main cause of <disease2 id="ID2">DISEASE</disease2>.',
						'<drug id="ID1">DRUG</drug> failed clinical trials for <disease2 id="ID2">DISEASE</disease2>.',
						'<disease2 id="ID2">DISEASE</disease2> is a known side effect of <drug id="ID1">DRUG</drug>.']
						
	fakeDrugNames = ['bmzvpvwbpw','pehhjnlvve''wbjccovflf','usckfljzxu','ruswdgzajr','vgypkemhjr','oxzbaapqct','elvptnpvyc']
	fakeDiseaseNames = ['gnorcyvmer','hfymprbifs','ootopaoxbg','knetvjnjun','kfjqxlpvew','zgwivlcmly','kneqlzjegs','kyekjnkrfo']
	
	totalCount = positiveCount + negativeCount
	
	relNames = [ "treats_%d" % i for i in range(relTypes) ]

	entityID = 1
	corpus = kindred.Corpus()
	for _ in range(positiveCount):
		text = customChoice(positivePatterns)
		text = text.replace('DRUG',customChoice(fakeDrugNames))
		text = text.replace('DISEASE',customChoice(fakeDiseaseNames))
		
		text = text.replace('ID1',str(entityID))
		text = text.replace('ID2',str(entityID+1))
		#relations = [ kindred.Relation('treats',[entityID,entityID+1]) ]

		relName = customChoice(relNames)
		text += '<relation type="%s" subj="%d" obj="%d" />' % (relName,entityID,entityID+1)
		
		entityID += 2
		
		converted = kindred.Document(text)
		corpus.addDocument(converted)
		
	halfNegativeCount = int(negativeCount/2.0)
	for _ in range(halfNegativeCount):
		combinedText = ""
		for _ in range(2):
			text = customChoice(negativePatterns)
			text = text.replace('DRUG',customChoice(fakeDrugNames))
			text = text.replace('DISEASE',customChoice(fakeDiseaseNames))
			
			text = text.replace('ID1',str(entityID))
			text = text.replace('ID2',str(entityID+1))
			entityID += 2
		
			combinedText = "%s %s" % (combinedText,text)
		
		converted = kindred.Document(combinedText)
		corpus.addDocument(converted)
		
	return corpus
	
def generateTestData(positiveCount = 100,negativeCount = 100, relTypes = 1):
	customSeed(b'seed')
	corpus = generateData(positiveCount, negativeCount, relTypes)
	docCount = len(corpus.documents)
	halfDataCount = int(docCount/2.0)
	trainIndices = customSample(range(docCount),halfDataCount)
	testIndices = [ i for i in range(docCount) if not i in trainIndices ]

	trainCorpus = kindred.Corpus()
	testCorpus = kindred.Corpus()
	for i in trainIndices:
		trainCorpus.addDocument(corpus.documents[i])
	for i in testIndices:
		testCorpus.addDocument(corpus.documents[i])

	return trainCorpus,testCorpus




import kindred

class Corpus:
	"""
	Collection of text documents.
	"""
	
	def __init__(self,text=None):
		"""
		Constructor
		
		:param text: Optional SimpleTag text to initalize a single document
		:type text: String (with SimpleTag format XML)
		"""

		self.documents = []
		if not text is None:
			doc = kindred.Document(text)
			self.addDocument(doc)
		self.parsed = False

		self.relationTypes = None

	def addDocument(self,doc):
		"""
		Add a single document to the corpus
		
		:param doc: Document to add
		:type doc: kindred.Document
		"""

		assert isinstance(doc,kindred.Document)
		self.documents.append(doc)


	def addRelationTypes(self,relationTypes):
		"""
		Add a set of relation types that have been identified in corpus
		
		:param relationTypes: List of relation type names
		:type relationTypes: List of strings
		"""

		self.relationTypes = relationTypes

	def clone(self):
		"""
		Clone the corpus
		
		:return: Clone of the corpus
		:rtype: kindred.Corpus
		"""

		cloned = Corpus()
		for doc in self.documents:
			cloned.addDocument(doc.clone())
		return cloned
	
	def getCandidateClasses(self):
		"""
		Get all the classes (i.e. indices of relation types) for all the candidate relations in this corpus.
		
		:return: List of indices (corresponding to the relation types) for each candidate relation. 0 means no relation type
		:rtype: List of integers
		"""

		classes = []
		for doc in self.documents:
			classes += doc.getCandidateClasses()
		return classes
		
	def getCandidateRelations(self):
		"""
		Get all the candidate relations in this corpus.
		
		:return: List of candidate relations
		:rtype: List of kindred.Relation
		"""

		relations = []
		for doc in self.documents:
			relations += doc.getCandidateRelations()
		return relations
		
	def getEntityMapping(self):
		"""
		Gets mapping from entity ID to entity instances
		
		:return: Map of entity ID to entity instance
		:rtype: dict
		"""

		entityMapping = {}
		for doc in self.documents:
			for e in doc.entities:
				entityMapping[e.entityID] = e
		return entityMapping

	def getRelations(self):
		"""
		Get all relations in this corpus
		
		:return: List of relations
		:rtype: list
		"""

		relations = []
		for doc in self.documents:
			relations += doc.getRelations()
		return relations

	def removeRelations(self):
		"""
		Remove all relations in this corpus
		"""

		for doc in self.documents:
			doc.removeRelations()



import codecs
import os
import json
from pprint import pprint
import sys
import re
from collections import OrderedDict

from xml.dom import minidom

import kindred

import bioc
import sys
from future.utils import native

def loadEntity(line,text):
	assert line[0] == 'T', "Entity input should start with a T"
	split = line.strip().split('\t')
	assert len(split) == 3
	entityID = split[0]
	typeInfo = split[1]
	tokens = split[2]
		
	textChunks = []
	typeSpacePos = typeInfo.index(' ')
	typeName = typeInfo[:typeSpacePos]
	positionText = typeInfo[typeSpacePos:]
	positions = []
	for coordinates in positionText.strip().split(';'):
		a,b = coordinates.strip().split(' ')
		a,b = int(a.strip()),int(b.strip())
		textChunk = text[a:b].replace('\n',' ').strip()
		textChunks.append(textChunk)
		positions.append((a,b))
		
	# Check that the tokens match up to the text
	chunkTest = " ".join(textChunks)
	tokensTest = tokens
	chunkTest = re.sub(r'\s\s+', ' ', chunkTest)
	tokensTest = re.sub(r'\s\s+', ' ', tokensTest)
	chunkTest = chunkTest.strip()
	tokensTest = tokensTest.strip()

	assert chunkTest == tokensTest , u"For id=" + entityID + ", tokens '" + tokens.encode('ascii', 'ignore') + "' don't match up with positions: " + str(positions)
	
	entity = kindred.Entity(typeName, tokensTest, positions, entityID)

	return entity
	
def loadRelation(line,ignoreComplexRelations=False):
	assert line[0] == 'E' or line[0] == 'R', "Relation input should start with a E or R"
	split = line.strip().split('\t')
	relationID = split[0]
	eventInfo = split[1]
	typeSpacePos = eventInfo.index(' ')
	

	eventNameSplit = eventInfo[:typeSpacePos].split(':')
	assert len(eventNameSplit) == 1, "Cannot load trigger events"
	relationType = eventNameSplit[0]
		
	isComplexRelation = False
	argumentText = eventInfo[typeSpacePos:]
	arguments = []
	for argument in argumentText.strip().split(' '):
		split2 = argument.strip().split(':')
		assert len(split2) == 2
		argName = split2[0]
		entityID = split2[1]

		isComplexRelation = (entityID[0] == 'R' or entityID[0] == 'E')

		# We'll skip this relation as
		if ignoreComplexRelations and isComplexRelation:
			break

		assert not isComplexRelation, "kindred does not support complex relations (where one relation has another relation as an argument), use ignoreComplexRelations=True to ignore these"

		assert not argName in arguments
		arguments.append((argName,entityID))

	if ignoreComplexRelations and isComplexRelation:
		return None

	arguments = sorted(arguments)
	entityIDs = [ entityID for argName,entityID in arguments ]
	argNames = [ argName for argName,entityID in arguments ]

	relation = kindred.Relation(relationType, entityIDs, argNames)
	return relation
	
# TODO: Deal with complex relations more clearly
def loadDataFromSTFormat(txtFile,a1File,a2File,verbose=False,ignoreEntities=[],ignoreComplexRelations=True):
	with codecs.open(txtFile, "r", "utf-8") as f:
		text = f.read()
			
	entities = []
	with codecs.open(a1File, "r", "utf-8") as f:
		for line in f:
			if line.strip() == '':
				continue
				
			assert line[0] == 'T', "Only triggers are expected in a1 file: " + a1File
			entity = loadEntity(line.strip(), text)
			if (not entity is None) and (not entity.entityType in ignoreEntities):
				entities.append(entity)
			
	relations = []
	if os.path.exists(a2File):
		with codecs.open(a2File, "r", "utf-8") as f:
			for line in f:
				if line.strip() == '':
					continue
					
				if line[0] == 'E' or line[0] == 'R':
					relation = loadRelation(line.strip(),ignoreComplexRelations)
					if not relation is None:
						relations.append(relation)
				elif verbose:
					sys.stderr.write("Unable to process line: %s\n" % line.strip())
	elif verbose:
		sys.stderr.write("Note: No A2 file found : %s\n" % os.path.basename(a2File))

	baseTxtFile = os.path.basename(txtFile)
	baseFilename = baseTxtFile[0:-4]
	combinedData = kindred.Document(text,entities=entities,relations=relations,sourceFilename=baseFilename)
			
	return combinedData

def parseJSON(data,ignoreEntities=[]):
	entities = []
	relations = []
	
	text = data['text']
	if 'denotations' in data:
		for d in data['denotations']:
			sourceEntityID = None
			if 'id' in d:
				sourceEntityID = d['id']
			
			entityType = d['obj']
			span = d['span']
			startPos,endPos = span['begin'],span['end']
			position = [(startPos,endPos)]
			entityText = text[startPos:endPos]
			
			if not entityType in ignoreEntities:
				entity = kindred.Entity(entityType,entityText,position,sourceEntityID=sourceEntityID)
				entities.append(entity)
	if 'relations' in data:
		for r in data['relations']:
			relationID = r['id']
			obj = r['obj']
			relationType = r['pred']
			subj = r['subj']
			
			entityIDs = [obj,subj]
			argNames = ['obj','subj']
			
			relation = kindred.Relation(relationType=relationType,entityIDs=entityIDs,argNames=argNames)
			relations.append(relation)
	#if 'modifications' in data:
	#	for m in data['modifications']:
	#		id = m['id']
	#		obj = m['obj']
	#		pred = m['pred']
	#		modification = (pred,obj)
	#		modifications[id] = modification

	#print "keys:", list(data.keys())
	expected = ['denotations','divid','modifications','namespaces','project','relations','sourcedb','sourceid','target','text','tracks']
	extraFields = [ k for k in data.keys() if not k in expected]
	assert len(extraFields) == 0, "Found additional unexpected fields (%s) in JSON" % (",".join(extraFields))
		
	combinedData = kindred.Document(text,entities=entities,relations=relations)

	return combinedData

def loadDataFromJSON(filename,ignoreEntities=[]):
	with open(filename) as f:
		data = json.load(f)
	parsed = parseJSON(data,ignoreEntities)
	
	baseTxtFile = os.path.basename(filename)
	parsed.sourceFilename = baseTxtFile
	
	return parsed
	
def parseSimpleTag_helper(node,currentPosition=0,ignoreEntities=[]):
	text,entities,relations = '',[],[]
	for s in node.childNodes:
		if s.nodeType == s.ELEMENT_NODE:
			insideText,insideEntities,insideRelations = parseSimpleTag_helper(s,currentPosition+len(text))

			if s.tagName == 'relation':
				relationType = s.getAttribute('type')
				arguments = [ (argName,entityID) for argName,entityID in s.attributes.items() if argName != 'type' ]
				arguments = sorted(arguments)
				
				entityIDs = [ entityID for argName,entityID in arguments]
				argNames = [ argName for argName,entityID in arguments]
				
				r = kindred.Relation(relationType=relationType,entityIDs=entityIDs,argNames=argNames)
				relations.append(r)
			else: # Entity
				entityType = s.tagName
				sourceEntityID = s.getAttribute('id')
				position = [(currentPosition+len(text),currentPosition+len(text)+len(insideText))]
				if not entityType in ignoreEntities:
					e = kindred.Entity(entityType,insideText,position,sourceEntityID=sourceEntityID)
					entities.append(e)
				
			text += insideText
			entities += insideEntities
			relations += insideRelations
		elif s.nodeType == s.TEXT_NODE:
			text += s.nodeValue
			
	return text,entities,relations
	
def mergeEntitiesWithMatchingIDs(unmergedEntities):
	assert isinstance(unmergedEntities,list)

	entityDict = OrderedDict()
	for e in unmergedEntities:
		assert isinstance(e, kindred.Entity)
		if e.sourceEntityID in entityDict:
			position = e.position
			entityDict[e.sourceEntityID].text += " " + e.text
			entityDict[e.sourceEntityID].position += e.position
		else:
			entityDict[e.sourceEntityID] = e
			
	return list(entityDict.values())
	
def parseSimpleTag(text,ignoreEntities=[]):
	docText = u"<doc>%s</doc>" % text
	xmldoc = minidom.parseString(docText.encode('utf8'))
	docNode = xmldoc.childNodes[0]
	text,unmergedEntities,relations = parseSimpleTag_helper(docNode,ignoreEntities=ignoreEntities)
	
	missingSourceEntityID = [ e.sourceEntityID == '' for e in unmergedEntities ]
	assert all(missingSourceEntityID) or (not any(missingSourceEntityID)), 'All entities or none (not some) should be given IDs'
	assert (not any(missingSourceEntityID)) or len(relations) == 0, "Cannot include relations with no-ID entities"
	
	if all(missingSourceEntityID):
		for i,e in enumerate(unmergedEntities):
			e.sourceEntityID = i+1
					
	entities = mergeEntitiesWithMatchingIDs(unmergedEntities)
			
	combinedData = kindred.Document(text,entities=entities,relations=relations)
	return combinedData

def loadDataFromBioC(filename,ignoreEntities=[]):
	with open(filename, 'r') as fp:
		collection = bioc.load(fp)
	
	parsed = []
	
	assert isinstance(collection,bioc.BioCCollection)
	
	for document in collection.documents:
		assert isinstance(document,bioc.BioCDocument)
		for passage in document.passages:
			assert isinstance(passage,bioc.BioCPassage)
			
			text = passage.text
			entities = []
			relations = []
			
			for a in passage.annotations:
				assert isinstance(a,bioc.BioCAnnotation)
				
				entityType = a.infons['type']
				sourceEntityID = a.id
				
				position = []
				segments = []
				
				for l in a.locations:
					assert isinstance(l,bioc.BioCLocation)
					startPos = int(native(l.offset))
					endPos = startPos + int(native(l.length))
					position.append((startPos,endPos))
					segments.append(text[startPos:endPos])
				
				entityText = " ".join(segments)
				e = kindred.Entity(entityType,entityText,position,sourceEntityID)
				entities.append(e)
				
			for r in passage.relations:
				assert isinstance(r,bioc.BioCRelation)
				relationID = r.id
				relationType = r.infons['type']
				
				arguments = []
				for n in r.nodes:
					assert isinstance(n,bioc.BioCNode)
					arguments.append((n.role,n.refid))
				arguments = sorted(arguments)
					
				entityIDs = [ entityID for argName,entityID in arguments]
				argNames = [ argName for argName,entityID in arguments]
				
				r = kindred.Relation(relationType=relationType,entityIDs=entityIDs,argNames=argNames)
				relations.append(r)
				
			relData = kindred.Document(text,entities=entities,relations=relations)
			parsed.append(relData)
			
	return parsed
	
	
def loadDoc(dataFormat,path=None,txtPath=None,a1Path=None,a2Path=None,verbose=False,ignoreEntities=[],ignoreComplexRelations=False):
	"""
	Loads a single document from a single file (for JSON or SimpleTag) or set of files (for standoff)
	"""
	
	assert dataFormat == 'standoff' or dataFormat == 'simpletag' or dataFormat == 'json'
	
	if dataFormat == 'standoff':
		assert not txtPath is None
		assert not a1Path is None
		#assert not a2Path is None

		doc = loadDataFromSTFormat(txtPath,a1Path,a2Path,verbose=verbose,ignoreEntities=ignoreEntities)
	elif dataFormat == 'simpletag':
		assert not path is None

		with open(path,'r') as f:
			filecontents = f.read().strip()
		doc = parseSimpleTag(filecontents,ignoreEntities=ignoreEntities)
		doc.sourceFilename = os.path.basename(path)
	elif dataFormat == 'json':
		assert not path is None
		doc = loadDataFromJSON(path,ignoreEntities=ignoreEntities)

	assert isinstance(doc,kindred.Document)
	return doc

def loadDocs(dataFormat,path=None,txtPath=None,a1Path=None,a2Path=None,verbose=False,ignoreEntities=[],ignoreComplexRelations=False):
	"""
	Loads a set of documents from a single file (for BioC)
	"""
	assert dataFormat == 'bioc'
	if dataFormat == 'bioc':
		assert not path is None
		docs = loadDataFromBioC(path,ignoreEntities=ignoreEntities)
		
	assert isinstance(docs,list)
	for doc in docs:
		assert isinstance(doc,kindred.Document)
		
	return docs
	
def loadDir(dataFormat,directory,verbose=False,ignoreEntities=[],ignoreComplexRelations=False):
	"""
	Loads a directory of files with the given data format.
	"""
	assert dataFormat == 'standoff' or dataFormat == 'simpletag' or dataFormat == 'json' or dataFormat == 'bioc'
	assert os.path.isdir(directory), "%s must be a directory"
	
	if directory[-1] != '/':
		directory += '/'

	filenames = sorted(list(os.listdir(directory)))

	corpus = kindred.Corpus()
	for filename in filenames:
		if dataFormat == 'standoff' and filename.endswith('.txt'):
			base = filename[0:-4]
			txtPath = os.path.join(directory, base + '.txt')
			a1Path = os.path.join(directory, base + '.a1')
			a2Path = os.path.join(directory, base + '.a2')

			assert os.path.isfile(txtPath), "%s must exist" % txtPath
			assert os.path.isfile(a1Path), "%s must exist" % a1Path

			doc = loadDoc(dataFormat,txtPath=txtPath,a1Path=a1Path,a2Path=a2Path,verbose=verbose,ignoreEntities=ignoreEntities,ignoreComplexRelations=ignoreComplexRelations)
			corpus.addDocument(doc)
		elif dataFormat == 'simpletag' and filename.endswith('.simple'):
			absPath = os.path.join(directory, filename)
			doc = loadDoc(dataFormat,path=absPath,verbose=verbose,ignoreEntities=ignoreEntities,ignoreComplexRelations=ignoreComplexRelations)
			corpus.addDocument(doc)
		elif dataFormat == 'json' and filename.endswith('.json'):
			absPath = os.path.join(directory, filename)
			doc = loadDoc(dataFormat,path=absPath,verbose=verbose,ignoreEntities=ignoreEntities,ignoreComplexRelations=ignoreComplexRelations)
			corpus.addDocument(doc)
		elif dataFormat == 'bioc' and filename.endswith('.bioc.xml'):
			absPath = os.path.join(directory, filename)
			docs = loadDocs(dataFormat,path=absPath,verbose=verbose,ignoreEntities=ignoreEntities,ignoreComplexRelations=ignoreComplexRelations)
			for doc in docs:
				corpus.addDocument(doc)
			
	return corpus
			
			




from collections import defaultdict
import itertools

import kindred

class CandidateBuilder:
	"""
	Generates set of all possible relations in corpus.
	"""
	def __init__(self):
		"""
		Constructor
		"""
		self.fitted = False

	def fit_transform(self,corpus):
		"""
		Creates the set of all possible relations that exist within the given corpus and adds these to the corpus under each kindred.Sentence instance. Each relation will be contained within a single sentence. This fitting function should be called the first time in order to initialise the set of known relationship types.
		
		:param corpus: Corpus of text with which to build relation candidates
		:type corpus: kindred.Corpus
		"""

		assert self.fitted == False, "CandidateBuilder has already been fit to corpus"
		assert isinstance(corpus,kindred.Corpus)

		if not corpus.parsed:
			parser = kindred.Parser()
			parser.parse(corpus)
		
		self.relTypes = set()
	
		for doc in corpus.documents:
			knownRelations = doc.getRelations()
			for r in knownRelations:
				assert isinstance(r,kindred.Relation)
			
			tmpRelTypesAndArgCount = [ tuple([r.relationType] + r.argNames) for r in knownRelations ]
			self.relTypes.update(tmpRelTypesAndArgCount)
			
		self.relTypes = sorted(list(self.relTypes))
		self.relClasses = { relType:(i+1) for i,relType in enumerate(self.relTypes) }
			
		self.fitted = True
	
		self.transform(corpus)

	def transform(self,corpus):
		"""
		Creates the set of all possible relations that exist within the given corpus and adds these to the corpus under each kindred.Sentence instance. Each relation will be contained within a single sentence.
		
		:param corpus: Corpus of text with which to build relation candidates
		:type corpus: kindred.Corpus
		"""
		assert self.fitted == True, "CandidateBuilder must be fit to corpus first"
		assert isinstance(corpus,kindred.Corpus)

		if not corpus.parsed:
			parser = kindred.Parser()
			parser.parse(corpus)

		for doc in corpus.documents:
			existingRelations = defaultdict(list)
			for r in doc.getRelations():
				assert isinstance(r,kindred.Relation)
				
				relationType = r.relationType
				entityIDs = tuple(r.entityIDs)
				
				relKey = tuple([r.relationType] + r.argNames)
				if relKey in self.relClasses:
					relationClass = self.relClasses[relKey]
					existingRelations[entityIDs].append(relationClass)

			for sentence in doc.sentences:
				entitiesInSentence = sentence.getEntityIDs()
							
				for entitiesInRelation in itertools.permutations(entitiesInSentence,2):
					candidateRelation = kindred.Relation(entityIDs=list(entitiesInRelation))
					candidateClass = [0]
					relKey = tuple(entitiesInRelation)
					if relKey in existingRelations:
						candidateClass = existingRelations[relKey]
					
					sentence.addCandidateRelation(candidateRelation,candidateClass)

				sentence.candidateRelationsProcessed = True
					
		corpus.addRelationTypes(self.relTypes)




"""
Importer for BioNLP Shared Task data
"""

import tempfile

import kindred


import zipfile
import hashlib
import os
import sys
import wget
import shutil
import six

if sys.version_info >= (3, 0):
	import urllib.request
else:
	import urllib

def _calcSHA256(filename):
	return hashlib.sha256(open(filename, 'rb').read()).hexdigest()

def _isDirEmpty(path):
	files = os.listdir(path)
	return files == []

def _findDir(name, path):
	assert os.path.isdir(path), "Must provide directory as path"
	for root, dirs, files in os.walk(path):
		if name in dirs:
			return os.path.abspath(os.path.join(root, name))
	return None

def _downloadFiles(files,downloadDirectory):
	assert _isDirEmpty(downloadDirectory)

	for url,shortName,expectedSHA256 in files:
		downloadedPath = os.path.join(downloadDirectory,shortName)

		if not os.path.isfile(downloadedPath):
			try:
				wget.download(url,out=downloadedPath,bar=None)
				
				downloadedSHA256 = _calcSHA256(downloadedPath)
				assert downloadedSHA256 == expectedSHA256, "SHA256 mismatch with downloaded file: %s" % shortName
				
				if shortName.endswith('.zip'):
					zip_ref = zipfile.ZipFile(downloadedPath, 'r')
					zip_ref.extractall(path=downloadDirectory)
					zip_ref.close()
			except:
				raise RuntimeError("Unable to download BioNLP ST files")


def load(taskName,ignoreEntities=[]):
	tempDir = tempfile.mkdtemp()

	taskOptions = {}

	# 2016-BB3-event-train
	files = [('http://2016.bionlp-st.org/tasks/bb2/BioNLP-ST-2016_BB-event_train.zip','BioNLP-ST-2016_BB-event_train.zip', '3b02adff92d8ba8814c9901f4af7681863569fe40cd0d87914258d48f989bb96')]
	expectedDir = 'BioNLP-ST-2016_BB-event_train'
	taskOptions['2016-BB3-event-train'] = (files,expectedDir,False)
	# 2016-BB3-event-dev
	files = [('http://2016.bionlp-st.org/tasks/bb2/BioNLP-ST-2016_BB-event_dev.zip','BioNLP-ST-2016_BB-event_dev.zip', '6d6555b425e42ecc4edf8ffc378439ba722c59922cdf6c5819959861ff683533')]
	expectedDir = 'BioNLP-ST-2016_BB-event_dev'
	taskOptions['2016-BB3-event-dev'] = (files,expectedDir,False)
	# 2016-BB3-event-test
	files = [('http://2016.bionlp-st.org/tasks/bb2/BioNLP-ST-2016_BB-event_test.zip','BioNLP-ST-2016_BB-event_test.zip', 'a1fce4657227bc55aaac5df72d2df615474e0d96a1518234d3e11b3af1e910e7')]
	expectedDir = 'BioNLP-ST-2016_BB-event_test'
	taskOptions['2016-BB3-event-test'] = (files,expectedDir,False)

	# 2016-SeeDev-binary-train
	files = [('http://2016.bionlp-st.org/tasks/seedev/BioNLP-ST-2016_SeeDev-binary_train.zip','BioNLP-ST-2016_SeeDev-binary_train.zip', 'ecca965ae09a31c5675cc2a62f238e0eb3a83c164a925a7e8984dd4d9dbcec84')]
	expectedDir = 'BioNLP-ST-2016_SeeDev-binary_train'
	taskOptions['2016-SeeDev-binary-train'] = (files,expectedDir,False)
	# 2016-SeeDev-binary-dev
	files = [('http://2016.bionlp-st.org/tasks/seedev/BioNLP-ST-2016_SeeDev-binary_dev.zip','BioNLP-ST-2016_SeeDev-binary_dev.zip', 'ea8e0f8bb1bc62982bb9a1a207df33a106c78b14aeb9a0701894943ec9262326')]
	expectedDir = 'BioNLP-ST-2016_SeeDev-binary_dev'
	taskOptions['2016-SeeDev-binary-dev'] = (files,expectedDir,False)
	# 2016-SeeDev-binary-test
	files = [('http://2016.bionlp-st.org/tasks/seedev/BioNLP-ST-2016_SeeDev-binary_test.zip','BioNLP-ST-2016_SeeDev-binary_test.zip', '744df7379a1777f7bdd92f37887f6ca03d15875bf8ae56389e409d5df9e614d2')]
	expectedDir = 'BioNLP-ST-2016_SeeDev-binary_test'
	taskOptions['2016-SeeDev-binary-test'] = (files,expectedDir,False)
	
	# 2016-SeeDev-full-train
	files = [('http://2016.bionlp-st.org/tasks/seedev/BioNLP-ST-2016_SeeDev-full_train.zip','BioNLP-ST-2016_SeeDev-full_train.zip', '52ac808251cf75a4c65f0227f53faf0d39dd6cb3a508becbce0a8dd131a2b7c4')]
	expectedDir = 'BioNLP-ST-2016_SeeDev-full_train'
	taskOptions['2016-SeeDev-full-train'] = (files,expectedDir,True)
	# 2016-SeeDev-full-dev
	files = [('http://2016.bionlp-st.org/tasks/seedev/BioNLP-ST-2016_SeeDev-full_dev.zip','BioNLP-ST-2016_SeeDev-full_dev.zip', '12a07592799197238a8b1b127b759e0d94dd30c52471538553006cafa2203638')]
	expectedDir = 'BioNLP-ST-2016_SeeDev-full_dev'
	taskOptions['2016-SeeDev-full-dev'] = (files,expectedDir,True)
	# 2016-SeeDev-full-test
	files = [('http://2016.bionlp-st.org/tasks/seedev/BioNLP-ST-2016_SeeDev-full_test.zip','BioNLP-ST-2016_SeeDev-full_test.zip', '253b45e1e0aae16f7ce19db4692633556d19fb9e5509e75dbb71de319fb7c2f5')]
	expectedDir = 'BioNLP-ST-2016_SeeDev-full_test'
	taskOptions['2016-SeeDev-full-test'] = (files,expectedDir,True)

	assert taskName in taskOptions.keys(), "%s not a valid option in %s" % (taskName, taskOptions.keys())
	filesToDownload,expectedDir,ignoreComplexRelations = taskOptions[taskName]

	try:
		_downloadFiles(filesToDownload,tempDir)
	except:
		exc_info = sys.exc_info()
		shutil.rmtree(tempDir)
		six.reraise(*exc_info)

	mainDir = _findDir(expectedDir,tempDir)

	corpus = kindred.loadDir(dataFormat='standoff',directory=mainDir,ignoreEntities=ignoreEntities,ignoreComplexRelations=ignoreComplexRelations)

	shutil.rmtree(tempDir)

	return corpus





import json

from kindred.pycorenlp import StanfordCoreNLP

import kindred
from intervaltree import Interval, IntervalTree
from collections import defaultdict
from kindred.Dependencies import initializeCoreNLP,checkCoreNLPDownload,testCoreNLPConnection

def shortenDepName(depName):
	acceptedSubNames = set(['acl:relcl','cc:preconj','compound:prt','det:predet','nmod:npmod','nmod:poss','nmod:tmod'])
	if depName in acceptedSubNames:
		return depName
	else:
	 	return depName.split(":")[0]

class Parser:
	"""
	Runs CoreNLP on corpus to get sentences and associated tokens
	"""
	
	def __init__(self,corenlp_url='http://localhost:9000'):
		self.corenlp_url = corenlp_url

	nlp = None

	def testConnection(self):
		if Parser.nlp is None:
			return False
	
		try:
			parsed = Parser.nlp.annotate("This is a test", properties={'annotators': 'tokenize,ssplit,lemma,pos,depparse,parse','outputFormat': 'json'})
			return True
		except:
			return False

	def setupConnection(self):
		if not checkCoreNLPDownload():
			raise RuntimeError("Cannot access local CoreNLP at http://localhost:9000 and cannot find CoreNLP files to launch subprocess. Please download using kindred.downloadCoreNLP() if subprocess should be used")

		initializeCoreNLP()
		Parser.nlp = StanfordCoreNLP(self.corenlp_url)
			
		assert self.testConnection() == True

	def parse(self,corpus):
		assert isinstance(corpus,kindred.Corpus)

		if self.testConnection() == False:
			self.setupConnection()
		
		for d in corpus.documents:
			entityIDsToEntities = d.getEntityIDsToEntities()
		
			denotationTree = IntervalTree()
			entityTypeLookup = {}
			for e in d.getEntities():
				entityTypeLookup[e.entityID] = e.entityType
			
				for a,b in e.position:
					denotationTree[a:b] = e.entityID
				
			parsed = Parser.nlp.annotate(d.getText(), properties={'annotators': 'tokenize,ssplit,lemma,pos,depparse,parse','outputFormat': 'json'})
	
			

			for sentence in parsed["sentences"]:
				#assert False
				tokens = []
				for t in sentence["tokens"]:
					#kindred.Token(token,pos,lemma,start,end)
					token = kindred.Token(t["word"],t["pos"],t["lemma"],t["characterOffsetBegin"],t["characterOffsetEnd"])
					tokens.append(token)

				dependencies = []
				for de in sentence["enhancedPlusPlusDependencies"]:
				#for de in sentence["collapsed-ccprocessed-dependencies"]:
					#depName = de["dep"].split(":")[0]
					depName = shortenDepName(de["dep"])
					#if depName == 'nmod:in_addition_to':
					#	assert False
					dep = (de["governor"]-1,de["dependent"]-1,depName)
					dependencies.append(dep)
				# TODO: Should I filter this more or just leave it for simplicity
					
				entityIDsToTokenLocs = defaultdict(list)
				for i,t in enumerate(tokens):
					entitiesOverlappingWithToken = denotationTree[t.startPos:t.endPos]
					for interval in entitiesOverlappingWithToken:
						entityID = interval.data
						entityIDsToTokenLocs[entityID].append(i)

				# Let's gather up the information about the "known" entities in the sentence
				entitiesWithLocations = []
				for entityID,entityLocs in sorted(entityIDsToTokenLocs.items()):
					entityType = entityTypeLookup[entityID]
					e = entityIDsToEntities[entityID]
					entityWithLocation = (e, entityLocs)
					entitiesWithLocations.append(entityWithLocation)
					
				sentence = kindred.Sentence(tokens, dependencies, entitiesWithLocations, d.getSourceFilename())
				d.addSentence(sentence)

		corpus.parsed = True




import kindred


class Document:
	"""
	Span of text with associated tagged entities and relations between entities.
	"""
	
	def __init__(self,text,entities=None,relations=None,relationsUseSourceIDs=True,sourceFilename=None):
		"""
		Constructor for a Document that can take text using the SimpleTag XML format, or a set of Entities and Relations with associated text.
		
		:param text: Text in document (plain-text, or SimpleTag)
		:param entities: Entities in document
		:param relations: Relations in document
		:param relationsUseSourceIDs: description
		:param sourceFilename: description
		:type text: type description
		:type entities: type description
		:type relations: type description
		:type relationsUseSourceIDs: type description
		:type sourceFilename: type description
		"""

		loadFromSimpleTag = (entities is None)

		self.sourceFilename = sourceFilename

		if loadFromSimpleTag:
			dataToCopy = kindred.loadFunctions.parseSimpleTag(text)
			self.text = dataToCopy.getText()
			self.entities = dataToCopy.getEntities()
			self.relations = dataToCopy.getRelations()
		else:
			self.text = text
			
			assert isinstance(entities,list)
			for e in entities:
				assert isinstance(e,kindred.Entity)
			self.entities = entities
			
			if relations is None:
				self.relations = []
			else:
				assert isinstance(relations,list)
				for r in relations:
					assert isinstance(r,kindred.Relation)
				self.relations = relations

		# We'll need to translate source IDs to internal IDs
		if relationsUseSourceIDs and not loadFromSimpleTag:
			sourceEntityIDsToEntityIDs = self.getSourceEntityIDsToEntityIDs()
			sourceEntityIDs = sourceEntityIDsToEntityIDs.keys()
			correctedRelations = []
			for r in self.relations:
				for e in r.entityIDs:
					assert e in sourceEntityIDs, "Entities in relation must occur in the associated text. %s does not" % e
				relationEntityIDs = [ sourceEntityIDsToEntityIDs[e] for e in r.entityIDs ]
				correctedR = kindred.Relation(r.relationType,relationEntityIDs,r.argNames)
				correctedRelations.append(correctedR)
				
			self.relations = correctedRelations

		self.sentences = []
		
	def __repr__(self):
		"""
		String representation of Document
		
		:return: string representation
		:rtype: str
		"""
		return self.__str__()
	
	def __str__(self):
		"""
		String representation of Document
		
		:return: string representation
		:rtype: str
		"""

		return u"<Document %s %s %s>"  % (self.text,str(self.entities),str(self.relations))
	
	def addRelation(self,relation):
		"""
		Add a relation to this document
		
		:param relation: Relation to add
		:type relation: kindred.Relation
		"""

		self.relations.append(relation)

	def addSentence(self,sentence):
		"""
		Add a sentence to this document
		
		:param sentence: Sentence to add
		:type sentence: kindred.Sentence
		"""

		assert isinstance(sentence,kindred.Sentence)
		self.sentences.append(sentence)
		
	def clone(self):
		"""
		Clones the document
		
		:return: Clone of the document
		:rtype: kindred.Document
		"""

		cloned = Document(self.text,entities=self.entities,relations=self.relations,relationsUseSourceIDs=False,sourceFilename=self.sourceFilename)
		return cloned

	def getCandidateClasses(self):
		"""
		Get all the classes (i.e. indices of relation types) for all the candidate relations in this document.
		
		:return: List of indices (corresponding to the relation types) for each candidate relation. 0 means no relation type
		:rtype: List of integers
		"""

		assert not self.sentences is None, "Document must be parsed and CandidateBuilder use to get candidate relations first"
		classes = []
		for sentence in self.sentences:
			assert sentence.candidateRelationsProcessed == True, "CandidateBuilder use to get candidate relations first"
			classes += [ relationtypeClass for relation,relationtypeClass in sentence.candidateRelationsWithClasses ]
		return classes
	
	def getCandidateRelations(self):
		"""
		Get all the candidate relations in this corpus.
		
		:return: List of candidate relations
		:rtype: List of kindred.Relation
		"""

		assert not self.sentences is None, "Document must be parsed and CandidateBuilder use to get candidate relations first"
		relations = []
		for sentence in self.sentences:
			assert sentence.candidateRelationsProcessed == True, "CandidateBuilder use to get candidate relations first"
			relations += [ relation for relation,relationtypeClass in sentence.candidateRelationsWithClasses ]
		return relations
		
	def getEntities(self):
		"""
		Get the entities for this document
		
		:return: List of entities
		:rtype: list of kindred.Entity
		"""
		
		return self.entities
	
	def getEntityIDs(self):
		"""
		Get the entity IDs for the entities in this document
		
		:return: List of entity IDs
		:rtype: list of int
		"""

		return [e.entityID for e in self.entities]
	
	def getEntityIDsToEntities(self):
		"""
		Get a mapping of entity IDs to entities
		
		:return: Map of entity ID to entity instance
		:rtype: dict
		"""

		return {e.entityID:e for e in self.entities}
		
	def getEntityIDsToEntityTypes(self):
		"""
		Get a mapping of entity IDs to entity types
		
		:return: Map of entity ID to entity types
		:rtype: dict
		"""

		return {e.entityID:e.entityType for e in self.entities}

	def getEntityIDsToSourceEntityIDs(self):
		"""
		Get a mapping of entity IDs to source entity IDs
		
		:return: Map of entity ID to source entity IDs
		:rtype: dict
		"""
		
		return {e.entityID:e.sourceEntityID for e in self.entities}
		
	def getRelations(self):
		"""
		Get the relations associated with the document
		
		:return: list of relations
		:rtype: list of kindred.Relation
		"""
		
		return self.relations
		
	def getSourceEntityIDsToEntityIDs(self):
		"""
		Get a mapping of source entity IDs to entity IDs
		
		:return: Map of source entity IDs to entity ID
		:rtype: dict
		"""
		
		return {e.sourceEntityID:e.entityID for e in self.entities}
		
	def getSourceFilename(self):
		"""
		Get the source filename for the document
		
		:return: Source filename
		:rtype: str
		"""
		
		return self.sourceFilename

	def getText(self):
		"""
		Get the text associated with the document
		
		:return: list of text
		:rtype: str
		"""
		
		return self.text

	def removeRelations(self):
		"""
		Remove all relations in this corpus
		"""
		self.relations = []





from collections import Counter,OrderedDict
from sklearn.feature_extraction import DictVectorizer
from sklearn.feature_extraction.text import TfidfTransformer
from scipy.sparse import coo_matrix, csr_matrix, lil_matrix, hstack, vstack

import kindred

class Vectorizer:
	"""
	Vectorizes set of candidate relations into scipy sparse matrix.
	"""
	
	def __init__(self,featureChoice=None,tfidf=True):
		self.fitted = False
		
		self._registerFunctions()
		validFeatures = self.featureInfo.keys()

		if featureChoice is None:
			self.chosenFeatures = validFeatures
		else:
			for f in featureChoice:
				assert f in validFeatures, "Feature (%s) is not a valid feature" % f
			self.chosenFeatures = featureChoice
		
		self.tfidf = tfidf


		self.dictVectorizers = {}
		self.tfidfTransformers = {}

	def _registerFunctions(self):
		self.featureInfo = OrderedDict()
		self.featureInfo['entityTypes'] = {'func':Vectorizer.doEntityTypes,'never_tfidf':True}
		self.featureInfo['unigramsBetweenEntities'] = {'func':Vectorizer.doUnigramsBetweenEntities,'never_tfidf':False}
		self.featureInfo['bigrams'] = {'func':Vectorizer.doBigrams,'never_tfidf':False}
		self.featureInfo['dependencyPathEdges'] = {'func':Vectorizer.doDependencyPathEdges,'never_tfidf':True}
		self.featureInfo['dependencyPathEdgesNearEntities'] = {'func':Vectorizer.doDependencyPathEdgesNearEntities,'never_tfidf':True}
		
	def getFeatureNames(self):
		assert self.fitted == True, "Must have fit data first"
		featureNames = []
		for feature in self.chosenFeatures:
			featureNames += self.dictVectorizers[feature].get_feature_names()
		return featureNames
		

	def doEntityTypes(self,corpus):
		entityMapping = corpus.getEntityMapping()
		data = []
		for cr in corpus.getCandidateRelations():
			tokenInfo = {}
			for argI,eID in enumerate(cr.entityIDs):
				eType = entityMapping[eID].entityType
				argName = "selectedtokentypes_%d_%s" % (argI,eType)
				tokenInfo[argName] = 1
			data.append(tokenInfo)
		return data
	
	def doUnigramsBetweenEntities(self,corpus):
		entityMapping = corpus.getEntityMapping()
		data = []	
		for doc in corpus.documents:
			for sentence in doc.sentences:
				for cr,_ in sentence.candidateRelationsWithClasses:
					dataForThisCR = Counter()

					assert len(cr.entityIDs) == 2
					pos1 = sentence.entityIDToLoc[cr.entityIDs[0]]
					pos2 = sentence.entityIDToLoc[cr.entityIDs[1]]
					
					if max(pos1) < min(pos2):
						startPos,endPos = max(pos1)+1,min(pos2)
					else:
						startPos,endPos = max(pos2)+1,min(pos1)

					tokenData = [ sentence.tokens[i].word.lower() for i in range(startPos,endPos) ]
					for t in tokenData:
						dataForThisCR[u"ngrams_betweenentities_%s" % t] += 1
					data.append(dataForThisCR)

		return data
	
	def doDependencyPathEdges(self,corpus):
		entityMapping = corpus.getEntityMapping()
		data = []	
		for doc in corpus.documents:
			for sentence in doc.sentences:
				for cr,_ in sentence.candidateRelationsWithClasses:
					dataForThisCR = Counter()

					assert len(cr.entityIDs) == 2
					pos1 = sentence.entityIDToLoc[cr.entityIDs[0]]
					pos2 = sentence.entityIDToLoc[cr.entityIDs[1]]

					combinedPos = pos1 + pos2

					nodes,edges = sentence.extractMinSubgraphContainingNodes(combinedPos)
					for a,b,dependencyType in edges:
						dataForThisCR[u"dependencypathelements_%s" % dependencyType] += 1
					data.append(dataForThisCR)

		return data
	
	def doDependencyPathEdgesNearEntities(self,corpus):
		entityMapping = corpus.getEntityMapping()
		data = []	
		for doc in corpus.documents:
			for sentence in doc.sentences:
				for cr,_ in sentence.candidateRelationsWithClasses:
					dataForThisCR = Counter()

					allEntityLocs = []
					for eID in cr.entityIDs:
						allEntityLocs += sentence.entityIDToLoc[eID]
					
					nodes,edges = sentence.extractMinSubgraphContainingNodes(allEntityLocs)
					for i,eID in enumerate(cr.entityIDs):

						pos = sentence.entityIDToLoc[eID]

						for a,b,dependencyType in edges:
							if a in pos:
								dataForThisCR[u"dependencypathnearselectedtoken_%d_%s" % (i,dependencyType)] += 1
					data.append(dataForThisCR)

		return data

	def doBigrams(self,corpus):
		entityMapping = corpus.getEntityMapping()
		data = []	
		for doc in corpus.documents:
			for sentence in doc.sentences:
				for cr,_ in sentence.candidateRelationsWithClasses:
					dataForThisCR = Counter()

					for _ in cr.entityIDs:

						startPos = 0
						endPos = len(sentence.tokens)

						tokenData = [ (sentence.tokens[i].word.lower(),sentence.tokens[i+1].word.lower()) for i in range(startPos,endPos-1) ]
						for t in tokenData:
							dataForThisCR[u"bigrams_%s_%s" % t] += 1
					data.append(dataForThisCR)

		return data

	def _vectorize(self,corpus,fit):
		assert isinstance(corpus,kindred.Corpus)
			
		matrices = []
		for feature in self.chosenFeatures:
			assert feature in self.featureInfo.keys()
			featureFunction = self.featureInfo[feature]['func']
			never_tfidf = self.featureInfo[feature]['never_tfidf']
			data = featureFunction(self,corpus)
			if fit:
				self.dictVectorizers[feature] = DictVectorizer()
				if self.tfidf and not never_tfidf:
					self.tfidfTransformers[feature] = TfidfTransformer()
					intermediate = self.dictVectorizers[feature].fit_transform(data)
					matrices.append(self.tfidfTransformers[feature].fit_transform(intermediate))
				else:
					matrices.append(self.dictVectorizers[feature].fit_transform(data))
			else:
				if self.tfidf and not never_tfidf:
					intermediate = self.dictVectorizers[feature].transform(data)
					matrices.append(self.tfidfTransformers[feature].transform(intermediate))
				else:
					matrices.append(self.dictVectorizers[feature].transform(data))

		mergedMatrix = hstack(matrices)
		return mergedMatrix
			
	def fit_transform(self,corpus):
		assert self.fitted == False
		self.fitted = True
		return self._vectorize(corpus,True)
	
	def transform(self,corpus):
		assert self.fitted == True
		return self._vectorize(corpus,False)
		
		
	



import kindred
import sys

class Entity:
	"""
	Biomedical entity with information of location in text
	"""
	
	nextInternalID = 1

	def __init__(self,entityType,text,position,sourceEntityID=None):
		posErrorMsg = "Entity position must be list of tuples (startPos,endPos)"

		if sys.version_info >= (3, 0):
			assert isinstance(entityType,str)
			assert isinstance(text,str)
		else:
			assert isinstance(entityType,basestring) or isinstance(entityType,unicode)
			assert isinstance(text,basestring) or isinstance(text,unicode)

		assert isinstance(position,list), posErrorMsg
		for p in position:
			assert isinstance(p,tuple), posErrorMsg
			assert len(p) == 2, posErrorMsg
			assert isinstance(p[0],int), posErrorMsg
			assert isinstance(p[1],int), posErrorMsg
	
		self.entityType = entityType
		self.sourceEntityID = sourceEntityID
		self.text = text
		self.position = position
		
		self.entityID = Entity.nextInternalID
		Entity.nextInternalID += 1
		
	def __str__(self):
		out = "<Entity %s:'%s' id=%d sourceid=%s %s>" % (self.entityType,self.text,self.entityID,str(self.sourceEntityID),str(self.position))
		return out
		
	def __repr__(self):
		return self.__str__()
		
	def __eq__(self, other):
		"""Override the default Equals behavior"""
		if isinstance(other, self.__class__):
			return self.__dict__ == other.__dict__
		return False
	
	def __ne__(self, other):
		"""Define a non-equality test"""
		return not self.__eq__(other)




"""
Importer for PubTator data
"""

import sys
import kindred
import requests
import re

from kindred.loadFunctions import parseJSON

def loadPMID(pmid):
	assert isinstance(pmid,int)
	
	annotationsURL = "https://www.ncbi.nlm.nih.gov/CBBresearch/Lu/Demo/RESTful/tmTool.cgi/BioConcept/%d/JSON" % pmid
	
	annotations = requests.get(annotationsURL).json()
	doc = parseJSON(annotations)
	
	return doc

def load(pmids):
	"""
	Load a set of documents with annotations from Pubmed given a list of Pubmed IDs (PMIDs)
	
	>>> corpus = load(19894120)
	>>> len(corpus.documents)
	1

	:param pmids: the list of Pubmed IDs (integers)
	:returns: a kindred corpus object
	"""

	assert isinstance(pmids,list) or isinstance(pmids,int)

	corpus = kindred.Corpus()
	if isinstance(pmids,list):
		for pmid in pmids:
			doc = loadPMID(pmid)
			assert isinstance(doc,kindred.Document)
			corpus.addDocument(doc)
	elif isinstance(pmids,int):
		doc = loadPMID(pmids)
		assert isinstance(doc,kindred.Document)
		corpus.addDocument(doc)
	return corpus
	
	
	
	



import kindred


class Token:
	"""
	Individual word with lemma, part-of-speech and location in text.
	"""
	
	def __init__(self,word,lemma,partofspeech,startPos,endPos):
		self.word = word
		self.lemma = lemma
		self.partofspeech = partofspeech
		self.startPos = startPos
		self.endPos = endPos

	def __str__(self):
		return self.word
		
	def __repr__(self):
		return self.__str__()




import zipfile
import hashlib
import os
import sys
import wget
import subprocess
import shlex
import time
import atexit
import tempfile
import requests
import pytest_socket

if sys.version_info >= (3, 0):
	import urllib.request
else:
	import urllib

homeDirectory = os.path.expanduser('~')
downloadDirectory = os.path.join(homeDirectory,'.kindred')
	
corenlpProcess = None
stdoutFile = None
stderrFile = None

def _calcSHA256(filename):
	return hashlib.sha256(open(filename, 'rb').read()).hexdigest()

def _findDir(name, path):
	for root, dirs, files in os.walk(path):
		if name in dirs:
			return os.path.abspath(os.path.join(root, name))
	return None

def _downloadFiles(files):
	global downloadDirectory
	
	if not os.path.isdir(downloadDirectory):
		os.mkdir(downloadDirectory)
	
	for url,shortName,expectedSHA256 in files:
		downloadedPath = os.path.join(downloadDirectory,shortName)
		
		# Check if the file is already downloaded (and delete if the SHA256 doesn't match)
		if os.path.isfile(downloadedPath):
			downloadedSHA256 = _calcSHA256(downloadedPath)
			if not downloadedSHA256 == expectedSHA256:
				os.remove(downloadedPath)

		if not os.path.isfile(downloadedPath):
			print("Downloading %s" % shortName)
			wget.download(url,out=downloadedPath,bar=None)
			
			downloadedSHA256 = _calcSHA256(downloadedPath)
			assert downloadedSHA256 == expectedSHA256
			
			if shortName.endswith('.zip'):
				print("Unzipping %s" % shortName)
				zip_ref = zipfile.ZipFile(downloadedPath, 'r')
				zip_ref.extractall(downloadDirectory)
				zip_ref.close()
			
def killCoreNLP():
	global corenlpProcess
	global stdoutFile
	global stderrFile
	if not corenlpProcess is None:
		corenlpProcess.kill()
		stdoutFile.close()
		stderrFile.close()

		corenlpProcess = None
		stdoutFile = None
		stderrFile = None

def checkCoreNLPDownload():
	directory = _findDir('stanford-corenlp-full-2016-10-31',downloadDirectory)
	return not directory is None

def downloadCoreNLP():
	directory = _findDir('stanford-corenlp-full-2016-10-31',downloadDirectory)
	if directory is None:
		files = []
		#files.append(('http://nlp.stanford.edu/software/stanford-corenlp-full-2016-10-31.zip','stanford-corenlp-full-2016-10-31.zip','753dd5aae1ea4ba14ed8eca46646aef06f6808a9ce569e52a09840f6928d00d8'))
		files.append(('http://bcgsc.ca/downloads/jlever/stanford-corenlp-full-2016-10-31.zip','stanford-corenlp-full-2016-10-31.zip','753dd5aae1ea4ba14ed8eca46646aef06f6808a9ce569e52a09840f6928d00d8'))
		
		print("Downloading CoreNLP to %s" % downloadDirectory)
		_downloadFiles(files)
		directory = _findDir('stanford-corenlp-full-2016-10-31',downloadDirectory)
		assert not directory is None, "Error after downloading, could not find corenlp directory"
		print ("Download complete.")
	else:
		print ("CoreNLP is already downloaded. No need to download")


def testCoreNLPConnection():
	try:
		requests.get('http://localhost:9000')
		return True
	except requests.exceptions.ConnectionError:
		return False
	except pytest_socket.SocketBlockedError:
		return False

def initializeCoreNLP():
	global corenlpProcess
	global stdoutFile
	global stderrFile

	if testCoreNLPConnection():
		return

	directory = _findDir('stanford-corenlp-full-2016-10-31',downloadDirectory)
	if directory is None:
		raise RuntimeError("Could not find the Stanford CoreNLP files. Use kindred.downloadCoreNLP() first")

	command='java -mx4g -cp "*" edu.stanford.nlp.pipeline.StanfordCoreNLPServer -port 9000 -timeout 150000'

	os.chdir(directory)

	stdoutFile = tempfile.NamedTemporaryFile(delete=True)
	stderrFile = tempfile.NamedTemporaryFile(delete=True)

	corenlpProcess = subprocess.Popen(shlex.split(command), stdout=stdoutFile, stderr=stderrFile, cwd=directory)#, shell=True)

	atexit.register(killCoreNLP)

	maxTries = 10

	connectionSuccess = False
	for tries in range(maxTries):
		if testCoreNLPConnection():
			connectionSuccess = True
			break
		time.sleep(5)

	if not connectionSuccess:
		killCoreNLP()
		raise RuntimeError("Unable to connect to launched CoreNLP subprocess")

	time.sleep(1)
		




import kindred
from collections import Counter

def evaluate(goldCorpus,testCorpus,metric='f1score',display=True):
	""" Compares the gold corpus with the test corpus and calculate appropriate metrics.
	
	:params goldCorpus: The gold standard set of data
	:type name: str
	:params testCorpus: The test set for comparison
	:type state: bool
	:params metric: Which metric to use (precision/recall/f1score)
	:type name: str
	:returns: float -- the score given the metric
	"""

	assert isinstance(goldCorpus,kindred.Corpus)
	assert isinstance(testCorpus,kindred.Corpus)

	TPs,FPs,FNs = Counter(),Counter(),Counter()
	
	#goldTuples = [ ]
	#for doc in goldCorpus.documents:
	#	relTuples = [ (r.relationType,tuple(r.entityIDs)) for r in doc.getRelations() ]
	#	goldTuples += relTuples
		
	goldTuples = [ (r.relationType,tuple(r.entityIDs)) for r in goldCorpus.getRelations() ]
	testTuples = [ (r.relationType,tuple(r.entityIDs)) for r in testCorpus.getRelations() ]

	totalSet = set(goldTuples + testTuples)
	for relation in totalSet:
		inGold = relation in goldTuples
		inTest = relation in testTuples

		relType = relation[0]
		
		if inGold and inTest:
			TPs[relType] += 1
		elif inGold:
			FNs[relType] += 1
		elif inTest:
			FPs[relType] += 1

	sortedRelTypes = sorted( list(set( [relation[0] for relation in totalSet] )))

	maxLen = max( [len(relType) for relType in sortedRelTypes ] )
	formatString = '%-' + str(maxLen) + 's\tTP:%d FP:%d FN:%d\tP:%f R:%f F1:%f'

	for relType in sortedRelTypes:
		TP,FP,FN = TPs[relType],FPs[relType],FNs[relType]
		precision = 0.0 if (TP+FP) == 0 else TP / float(TP+FP)
		recall = 0.0 if (TP+FN) == 0 else TP / float(TP+FN)
		f1score = 0.0 if precision==0 or recall == 0 else 2 * (precision*recall) / (precision+recall)
	
		if display:
			print(formatString % (relType,TP,FP,FN,precision,recall,f1score))

	TP,FP,FN = sum(TPs.values()),sum(FPs.values()),sum(FNs.values())
	precision = 0.0 if (TP+FP) == 0 else TP / float(TP+FP)
	recall = 0.0 if (TP+FN) == 0 else TP / float(TP+FN)
	f1score = 0.0 if precision==0 or recall == 0 else 2 * (precision*recall) / (precision+recall)

	if display:
		print("-"*50)
		print(formatString % ("All",TP,FP,FN,precision,recall,f1score))


	
	if metric == 'f1score':
		return f1score
	elif metric == 'precision':
		return precision
	elif metric == 'recall':
		return recall
	elif metric == 'all':
		return precision,recall,f1score	
	else:
		raise RuntimeError('Unknown metric: %s' % metric)




import kindred
import networkx as nx
import itertools
import sys

class Sentence:
	"""
	Set of tokens for a sentence after parsing
	"""
	
	def extractMinSubgraphContainingNodes(self, minSet):
		assert isinstance(minSet, list)
		for i in minSet:
			assert isinstance(i, int)
			assert i >= 0
			assert i < len(self.tokens)
		G1 = nx.Graph()
		for a,b,depType in self.dependencies:
			G1.add_edge(a,b,dependencyType=depType)

		G2 = nx.Graph()
		paths = {}

		minSet = sorted(list(set(minSet)))
		setCount1 = len(minSet)
		minSet = [ a for a in minSet if G1.has_node(a) ]
		setCount2 = len(minSet)
		if setCount1 != setCount2:
			sys.stderr.write("WARNING. %d node(s) not found in dependency graph!\n" % (setCount1-setCount2))
		for a,b in itertools.combinations(minSet,2):
			try:
				path = nx.shortest_path(G1,a,b)
				paths[(a,b)] = path
				G2.add_edge(a,b,weight=len(path))
			except nx.exception.NetworkXNoPath:
				sys.stderr.write("WARNING. No path found between nodes %d and %d!\n" % (a,b))
			
		# TODO: This may through an error if G2 ends up having multiple components. Catch it gracefully.
		minTree = nx.minimum_spanning_tree(G2)
		nodes = set()
		allEdges = set()
		for a,b in minTree.edges():
			path = paths[(min(a,b),max(a,b))]
			for i in range(len(path)-1):
				a,b = path[i],path[i+1]
				dependencyType = G1.get_edge_data(a,b)['dependencyType']
				edge = (min(a,b),max(a,b),dependencyType)
				allEdges.add(edge)
			nodes.update(path)

		return nodes,allEdges
	
	def __str__(self):
		tokenWords = [ t.word for t in self.tokens ]
		return " ".join(tokenWords)
	
	def __repr__(self):
		return self.__str__()

	def getEntityIDs(self):
		return [ e.entityID for e,_ in self.entitiesWithLocations ]
		
	def getEntityType(self,entityID):
		return self.entityIDToType[entityID]

	def addCandidateRelation(self, relation, relationtypeClass):
		self.candidateRelationsWithClasses.append((relation,relationtypeClass))

	def __init__(self, tokens, dependencies, entitiesWithLocations, sourceFilename=None):
		assert isinstance(tokens, list) 
		assert isinstance(dependencies, list) 
		assert isinstance(entitiesWithLocations, list)
		for entityWithLocation in entitiesWithLocations:
			assert isinstance(entityWithLocation,tuple)
			assert len(entityWithLocation) == 2
			assert isinstance(entityWithLocation[0],kindred.Entity)
			assert isinstance(entityWithLocation[1],list)
		
		self.tokens = tokens
		self.entitiesWithLocations = entitiesWithLocations
		self.sourceFilename = sourceFilename
		
		self.dependencies = dependencies
		
		self.entityIDToType = { e.entityID:e.entityType for e,_ in self.entitiesWithLocations }
		self.entityIDToLoc = { e.entityID:loc for e,loc in self.entitiesWithLocations }

		self.candidateRelationsWithClasses = []
		self.candidateRelationsProcessed = False
	





import numpy as np
from sklearn.linear_model import LogisticRegression

class LogisticRegressionWithThreshold:
	def __init__(self,threshold=0.5):
		#self.clf = svm.SVC(kernel='linear', class_weight='balanced', probability=True)
		self.clf = LogisticRegression(class_weight='balanced',random_state=1)
		self.threshold = threshold

	def fit(self,X,Y):
		self.clf.fit(X,Y)
		self.classes_ = self.clf.classes_

	def predict(self,X):
		probs = self.clf.predict_proba(X)

		# Ignore probabilities that fall below our threshold
		probs[probs<self.threshold] = -1.0

		# But make sure that the negative class (class=0) always has a slightly higher value
		probs[:,0][probs[:,0]<self.threshold] = -0.5

		# And get the highest probability for each row
		predictions = np.argmax(probs,axis=1)

		return predictions
		



"""
Importer for PubAnnotation data
"""

import sys
import kindred
import requests
import re

from kindred.loadFunctions import parseJSON

def load(projectName):
	projectURL = "http://pubannotation.org/projects/%s/docs.json" % projectName
	
	loaded = kindred.Corpus()
	
	docs = requests.get(projectURL)
	for doc in docs.json():
		m = re.search("sourcedb/(?P<sourcedb>[^\/]*)/sourceid/(?P<sourceid>[0-9]*)",doc['url'])
		mDict = m.groupdict()
		
		assert 'sourcedb' in mDict
		assert 'sourceid' in mDict
		
		annotationsURL = "http://pubannotation.org/projects/%s/docs/sourcedb/%s/sourceid/%s/annotations.json" % (projectName,mDict['sourcedb'],mDict['sourceid'])
		
		annotations = requests.get(annotationsURL).json()
		
		
		assert isinstance(annotations,list) or isinstance(annotations,dict)
		
		if isinstance(annotations,list):
			for annotation in annotations:
				parsed = parseJSON(annotation)
				loaded.addDocument(parsed)
		elif isinstance(annotations,dict):
			parsed = parseJSON(annotations)
			loaded.addDocument(parsed)
	
	return loaded
	
	
	



#!/usr/bin/env python

from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

def readme():
	with open('README.rst') as f:
		return f.read()

setup(name='kindred',
	version='1.0.0',
	description='A relation extraction toolkit for biomedical text mining',
	long_description=long_description,
	classifiers=[
		'Intended Audience :: Developers',
		'Intended Audience :: Education',
		'Intended Audience :: Information Technology',
		'Intended Audience :: Science/Research',
		'License :: OSI Approved :: MIT License',
		'Operating System :: OS Independent',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
		'Topic :: Scientific/Engineering',
		'Topic :: Scientific/Engineering :: Artificial Intelligence',
		'Topic :: Scientific/Engineering :: Human Machine Interfaces',
		'Topic :: Scientific/Engineering :: Information Analysis',
		'Topic :: Text Processing',
		'Topic :: Text Processing :: General',
		'Topic :: Text Processing :: Indexing',
		'Topic :: Text Processing :: Linguistic',
	],
	url='http://github.com/jakelever/kindred',
	author='Jake Lever',
	author_email='jake.lever@gmail.com',
	license='MIT',
	packages=['kindred'],
	install_requires=['scikit-learn','numpy','scipy','intervaltree','networkx','wget','pycorenlp','lxml','future','bioc','pytest_socket','six'],
	include_package_data=True,
	zip_safe=False,
	test_suite='nose.collector',
	tests_require=['nose'])





# Data types
from kindred.Corpus import Corpus
from kindred.Document import Document
from kindred.Entity import Entity
from kindred.Relation import Relation
from kindred.Token import Token
from kindred.Sentence import Sentence

# Components
from kindred.Parser import Parser
from kindred.CandidateBuilder import CandidateBuilder
from kindred.Vectorizer import Vectorizer
from kindred.RelationClassifier import RelationClassifier
from kindred.LogisticRegressionWithThreshold import LogisticRegressionWithThreshold

# General functions
from kindred.loadFunctions import loadDoc,loadDocs,loadDir
from kindred.saveFunctions import save
from kindred.evalFunctions import evaluate

# Data sources
from kindred import bionlpst, pubannotation, pubtator

# Installing
from kindred.Dependencies import downloadCoreNLP



		

		
	







# Temporary inclusion of pycorenlp code for easier edits
# https://github.com/smilli/py-corenlp

import json, requests

class StanfordCoreNLP:

    def __init__(self, server_url):
        self.server_url = server_url

    def annotate(self, text, properties={}):
        assert isinstance(text, str) or isinstance(text,unicode)
        assert isinstance(properties, dict)

        data = text.encode('utf8')
        r = requests.post(
            self.server_url, params={
                'properties': str(properties)
            }, data=data, headers={'Connection': 'close'})
        
        assert 'outputFormat' in properties and properties['outputFormat'] == 'json'
        output = json.loads(r.text, encoding='utf-8', strict=False)

        return output

